/* Copyright (C) 2020 Terence Kelly.  All rights reserved.
   Convert edge list representation of directed graph to
   "compressed sparse row" (CSR) format.  Input file is
   read repeatedly; it must not change.  Compile as C11. */

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define ERRSTR strerror(errno)
#define                           S1(s) #s
#define                     S2(s) S1(s)
#define COORDS __FILE__ ":" S2(__LINE__) ": "
#define BAIL(...)                           \
  do { fprintf(stderr, COORDS __VA_ARGS__); \
       exit(EXIT_FAILURE); } while (0)
#define CAL(p, n, s)                                      \
  do { if (NULL == ((p) = (int *)calloc((n), (s))))       \
         BAIL("calloc(%lu, %lu): %s\n", (n), (s), ERRSTR);\
     } while (0)

static struct {
  int V,  // max vertexID; valid vertexIDs are [1..V]
      E,  // total number of edges
     *N,  // indexed by "from" ID; outdeg(v) == N[1+v]-N[v]
     *F;  // "to" vertexIDs accessed via N[]
} CSR;

static void print_adjacencies(void) {
  printf("per-vertex adjacencies:\n");
  for (int a = 1; a <= CSR.V; a++) {
    int *begin = CSR.F + CSR.N[  a],
        *end   = CSR.F + CSR.N[1+a];
    printf("%d:", a);
    for (int *b = begin; b < end; b++)
      printf(" %d", *b);
    printf("\n");
  }
}

static_assert(sizeof(intmax_t) > sizeof(int), "int sizes");
static int s2i(const char *s) {
  char *p;  intmax_t r;
  errno = 0;
  r = strtoimax(s, &p, 10);
  if (0 != errno || '\0' != *p || 0 >= r || INT_MAX <= r)
    BAIL("s2i(\"%s\") -> %" PRIdMAX ", errno => %s\n",
         s, r, ERRSTR);
  return (int)r;
}

static int icmp(const void *a, const void *b) {
  return *(const int *)a - *(const int *)b;
}

int main(int argc, char *argv[]) {
  int a, b, line = 0, t = 0;
  FILE *fp;

  if (4 != argc)
    BAIL("usage:  %s  V  E  edgelistfile\n", argv[0]);
  CSR.V = s2i(argv[1]);
  CSR.E = s2i(argv[2]);
  if (NULL == (fp = fopen(argv[3], "r")))
    BAIL("fopen(\"%s\"): %s\n", argv[3], ERRSTR);

  CAL(CSR.N, 2 + (size_t)CSR.V, sizeof *CSR.N);
  CAL(CSR.F,     (size_t)CSR.E, sizeof *CSR.F);

  while (2 == fscanf(fp, "%d %d\n", &a, &b)) {
    line++;
    if (0 >= a || a > CSR.V || 0 >= b || b > CSR.V)
      BAIL("%d:  bad vertexID:  %d %d\n", line, a, b);
    if (a == b)
      fprintf(stderr, "%d:  warning:  self edge\n", line);
    CSR.N[a]++;
  }
  if (! feof(fp))
    BAIL("parse error after %d lines: %s\n", line, ERRSTR);
  if (line != CSR.E)
    BAIL("%d input lines != %d edges\n", line, CSR.E);

  for (a = 1; a <= CSR.V; a++) {
    t += CSR.N[a];
    CSR.N[a] = t;
  }
  CSR.N[a] = t;
  assert(CSR.N[1 + CSR.V] == CSR.E);

  rewind(fp);
  while (2 == fscanf(fp, "%d %d\n", &a, &b))
    CSR.F[--CSR.N[a]] = b;  // add directed edge a -> b

  if (0 != fclose(fp))
    BAIL("fclose(): %s\n", ERRSTR);

  for (a = 1; a <= CSR.V; a++)
    qsort(CSR.F + CSR.N[a],
          (size_t)(CSR.N[1+a] - CSR.N[a]),
          sizeof *CSR.F, icmp);

  printf("dump CSR format:\n"
         "V = %d    E = %d\n"
         "N: ", CSR.V, CSR.E);
  for (a = 0; a <= 1 + CSR.V; a++)
    printf(" %d", CSR.N[a]);
  printf("\n"
         "F: ");
  for (a = 0; a < CSR.E; a++)
    printf(" %d", CSR.F[a]);
  printf("\n");

  print_adjacencies();

  /* Intended for tests to verify that output emitted
     below is byte-for-byte identical to (sorted
     version of) input edge list, assuming input has
     no duplicate edges. */
  #ifdef TEST
  #define TESTFILE "out_edge_list.txt"
  if (NULL == (fp = fopen(TESTFILE, "w")))
    BAIL("fopen(\"%s\"): %s\n", TESTFILE, ERRSTR);
  for (a = 1; a <= CSR.V; a++) {
    int *begin = CSR.F + CSR.N[  a],
        *end   = CSR.F + CSR.N[1+a];
    for (int *b = begin; b < end; b++)
      fprintf(fp, "%d %d\n", a, *b);
  }
  if (0 != fclose(fp))
    BAIL("fclose(): %s\n", ERRSTR);
  #endif

  free(CSR.N);
  free(CSR.F);

  return 0;
}

